PREFERENCE_PHRASE_PATIENT = {
    'asap': 'You want the earliest available doctor in the department for the outpatient visit.',
    'doctor': 'You have a preferred doctor for the outpatient visit.',
    'date': 'You want the earliest available doctor in the department for the outpatient visit, starting from **{date}**.'
}
PREFERENCE_PHRASE_STAFF = {
    'asap': 'The patient wants the earliest available doctor in the department for the outpatient visit.',
    'doctor': 'The patient has a preferred doctor for the outpatient visit.',
    'date': 'The patient wants the earliest available doctor in the department for the outpatient visit, starting from **{date}**.'
}